import time

from mindsdb_sql.parser.ast import Identifier, Select, Insert

from mindsdb.utilities import log
from mindsdb.utilities.context import context as ctx

from .types import ChatBotMessage, BotException


class BasePolling:
    def __init__(self, chat_task, chat_params):
        self.params = chat_params
        self.chat_task = chat_task

    def start(self, stop_event):
        raise NotImplementedError

    def send_message(self, message: ChatBotMessage):
        chat_id = message.destination
        text = message.text

        t_params = self.params['chat_table']
        ast_query = Insert(
            table=Identifier(t_params['name']),
            columns=[t_params['chat_id_col'], t_params['text_col']],
            values=[
                [chat_id, text],
            ]
        )

        self.chat_task.chat_handler.query(ast_query)


class MessageCountPolling(BasePolling):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._to_stop = False
        self.chats_prev = None

    def run(self, stop_event):

        while True:
            try:
                chat_ids = self.check_message_count()
                for chat_id in chat_ids:
                    chat_memory = self.chat_task.memory.get_chat(chat_id)

                    message = self.get_last_message(chat_memory)
                    if message:
                        self.chat_task.on_message(chat_memory, message)

            except Exception as e:
                log.logger.error(e)

            if stop_event.is_set():
                return
            log.logger.debug(f'running {self.chat_task.bot_id}')
            time.sleep(7)

    def get_last_message(self, chat_memory):
        # retrive from history
        history = chat_memory.get_history()
        last_message = history[-1]
        if last_message.user == self.chat_task.bot_params['bot_username']:
            # the last message is from bot
            return
        return last_message

    def check_message_count(self):
        p_params = self.params['polling']

        chat_ids = []

        id_col = p_params['chat_id_col']
        msgs_col = p_params['count_col']
        # get chats status info
        ast_query = Select(
            targets=[
                Identifier(id_col),
                Identifier(msgs_col)],
            from_table=Identifier(p_params['table'])
        )

        resp = self.chat_task.chat_handler.query(query=ast_query)
        if resp.data_frame is None:
            raise BotException('Error to get count of messages')

        chats = {}
        for row in resp.data_frame.to_dict('records'):
            chat_id = row[id_col]
            msgs = row[msgs_col]

            chats[chat_id] = msgs

        if self.chats_prev is None:
            # first run
            self.chats_prev = chats
        else:
            # compare
            # for new keys
            for chat_id, count_msgs in chats.items():
                if self.chats_prev.get(chat_id) != count_msgs:
                    chat_ids.append(chat_id)

            self.chats_prev = chats
        return chat_ids

    def stop(self):
        self._to_stop = True


class RealtimePolling(BasePolling):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # call back can be without context
        self._ctx_dump = ctx.dump()

    def _callback(self, row, key):
        ctx.load(self._ctx_dump)

        row.update(key)

        t_params = self.params['chat_table']

        message = ChatBotMessage(
            ChatBotMessage.Type.DIRECT,
            row[t_params['text_col']],
            # In Slack direct messages are treated as channels themselves.
            row[t_params['username_col']],
            row[t_params['chat_id_col']]
        )

        chat_id = row[t_params['chat_id_col']]

        chat_memory = self.chat_task.memory.get_chat(chat_id)
        self.chat_task.on_message(chat_memory, message)

    def run(self, stop_event):
        t_params = self.params['chat_table']
        self.chat_task.chat_handler.subscribe(stop_event, self._callback, t_params['name'])

    # def send_message(self, message: ChatBotMessage):
    #
    #     self.chat_task.chat_handler.realtime_send(message)



import os
import random
import string
from typing import Optional

import numpy as np
import pandas as pd
import autokeras as ak
from sklearn import preprocessing
from tensorflow.keras.models import load_model

from mindsdb.integrations.libs.base import BaseMLEngine


# Makes this run on systems where this arg isn't specified, like Windows Subsystem for Linux
# Doesn't break things on Mac
os.environ["XLA_FLAGS"] = "--xla_gpu_cuda_data_dir=/usr/lib/cuda"

trainer_dict = {"regression": ak.StructuredDataRegressor, "classification": ak.StructuredDataClassifier}

DEFAULT_TRIALS = 100


def train_model(df, target, max_trials=DEFAULT_TRIALS):
    """Helper function to trains an AutoKeras model with an input df.

    Automatically decides on classification vs. regression depending on
    the type of the target.

    Will auto-encode categorical variables as dummies.

    Returns both trained model and the names of categoric dummy columns, which
    are passed later to the prediction method.
    """
    # Choose regressor of classifier based on target data type
    if np.issubdtype(df[target].dtype, np.number):
        mode = "regression"
        y_train = df[target]
    else:
        mode = "classification"
        lb = preprocessing.LabelBinarizer()
        y_train = lb.fit_transform(df[target])

    training_df = df.drop(target, axis=1)
    trainer = trainer_dict[mode](overwrite=True, max_trials=max_trials)

    # Save the column names of all numeric columns before transforming any categorical columns into dummies
    numeric_column_names = training_df.select_dtypes(include=[np.number]).columns.values.tolist()
    training_df = pd.get_dummies(training_df)
    categorical_dummy_column_names = [
        col for col in training_df.columns.values.tolist() if col not in numeric_column_names
    ]
    trainer.fit(training_df, y_train, verbose=2)
    return trainer.export_model(), categorical_dummy_column_names


def get_preds_from_model(df, model, target, column_count, categorical_dummy_column_names):
    """Gets predictions from the stored AutoKeras model."""
    df_to_predict = df.copy()
    for col in ["__mindsdb_row_id", target]:
        if col in df_to_predict.columns.values.tolist():
            df_to_predict = df_to_predict.drop(col, axis=1)

    if len(df_to_predict.columns) != column_count:
        raise Exception("All feature columns must be defined in the WHERE clause when making predictions")
    # Get dummies for any categorical columns and then populate the missing ones with zeros
    df_with_dummies = pd.get_dummies(df_to_predict)
    for col in categorical_dummy_column_names:  # exception handler for empty columns
        if col not in df_with_dummies.columns.values.tolist():
            df_with_dummies[col] = 0

    return model.predict(df_with_dummies, verbose=2)


def format_categorical_preds(predictions, original_y, df_to_predict, target_col):
    """Transforms class predictions back to their original class.

    Categoric predictions come out the AutoKeras model in a binary
    format e.g. (0, 1, 0). This function maps them back to their
    original class e.g. 'Blue', and adds a DF column for the
    model confidence score.
    """
    # Turn prediction back into categorical value
    lb = preprocessing.LabelBinarizer()
    lb.fit(original_y)
    preds = lb.inverse_transform(predictions)

    # Add the confidence score next to the prediction
    df_to_predict[target_col] = pd.Series(preds).astype(original_y.dtype)
    df_to_predict["confidence"] = [max(row) for _, row in enumerate(predictions)]
    return df_to_predict


class AutokerasHandler(BaseMLEngine):
    """
    Integration with the AutoKeras ML library.
    """  # noqa

    name = "autokeras"

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[dict] = None) -> None:
        """Create and tune AutoKeras model using the input df.

        Saves the AutoKeras best model params to model storage.
        """
        args = args["using"]  # ignore the rest of the problem definition
        args["target"] = target
        max_trials = int(args["train_time"] * DEFAULT_TRIALS) if "train_time" in args else DEFAULT_TRIALS
        # Save the training df in order to filter the training data based on the predict df
        args["training_df"] = df.to_json()
        args["training_data_column_count"] = len(df.columns) - 1  # subtract 1 for target

        random_string = "".join(random.choices(string.ascii_uppercase + string.digits, k=24))
        args["folder_path"] = os.path.join("autokeras", random_string)

        model, args["data_column_names"] = train_model(df, target, max_trials)
        model.save(args["folder_path"])
        self.model_storage.json_set("predict_args", args)

    def predict(self, df, args=None):
        """Predicts with best saved AutoKeras model."""
        args = self.model_storage.json_get("predict_args")
        training_df = pd.read_json(args["training_df"])
        model = load_model(args["folder_path"], custom_objects=ak.CUSTOM_OBJECTS)

        df_to_predict = df.copy()
        predictions = get_preds_from_model(df_to_predict, model, args["target"], args["training_data_column_count"], args["data_column_names"])

        # If we used the classifier we need to pre-process the predictions before returning them
        original_y = training_df[args["target"]]
        if not np.issubdtype(original_y.dtype, np.number):
            return format_categorical_preds(predictions, original_y, df_to_predict, args["target"])

        df_to_predict[args["target"]] = predictions
        return df_to_predict



import pandas as pd

from typing import List

from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb.utilities.log import get_log

from mindsdb_sql.parser import ast

logger = get_log("integrations.confluence_handler")

class ConfluenceSpacesTable(APITable):
    """Confluence Spaces Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the Confluence "get_all_spaces" API endpoint
        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query
        Returns
        -------
        pd.DataFrame
            confluence "get_all_spaces" matching the query
        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        conditions = extract_comparison_conditions(query.where)

        if query.limit:
            total_results = query.limit.value
        else:
            total_results = 50


        spaces_kwargs = {}
        order_by_conditions = {}

        if query.order_by and len(query.order_by) > 0:
            order_by_conditions["columns"] = []
            order_by_conditions["ascending"] = []

            for an_order in query.order_by:
                if an_order.field.parts[0] != "":
                    next    
                if an_order.field.parts[1] in self.get_columns():
                    order_by_conditions["columns"].append(an_order.field.parts[1])

                    if an_order.direction == "ASC":
                        order_by_conditions["ascending"].append(True)
                    else:
                        order_by_conditions["ascending"].append(False)
                else:
                    raise ValueError(
                        f"Order by unknown column {an_order.field.parts[1]}"
                    )

        for a_where in conditions:
            if a_where[1] == "type":
                if a_where[0] != "=":
                    raise ValueError("Unsupported where operation for type")
                if a_where[2] not in ["personal", "global"]:
                    raise ValueError(
                        f"Unsupported where argument for state {a_where[2]}"
                    )
                spaces_kwargs["type"] = a_where[2]
            else:
                raise ValueError(f"Unsupported where argument {a_where[1]}")

        confluence_spaces_records = self.handler.connect().get_all_spaces(start=0,limit=total_results)
        confluence_spaces_df = pd.json_normalize(confluence_spaces_records["results"])
        confluence_spaces_df = confluence_spaces_df[self.get_columns()]

        if "type" in spaces_kwargs:
            confluence_spaces_df = confluence_spaces_df[confluence_spaces_df.type == spaces_kwargs["type"]]
        
        selected_columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")


        if len(confluence_spaces_df) == 0:
            confluence_spaces_df = pd.DataFrame([], columns=selected_columns)
        else:
            confluence_spaces_df.columns = self.get_columns()
            for col in set(confluence_spaces_df.columns).difference(set(selected_columns)):
                confluence_spaces_df = confluence_spaces_df.drop(col, axis=1)

            if len(order_by_conditions.get("columns", [])) > 0:
                confluence_spaces_df = confluence_spaces_df.sort_values(
                    by=order_by_conditions["columns"],
                    ascending=order_by_conditions["ascending"],
                )

        return confluence_spaces_df

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses
        Returns
        -------
        List[str]
            List of columns
        """
        return [
            "id",
            "key",
            "name",
            "type",
            "_links.self",
            "_links.webui",
        ] 



import re
import os
from typing import Optional, Dict

import numpy as np
import pandas as pd

from langchain.schema import SystemMessage
from langchain.agents import AgentType
from langchain.llms import OpenAI
from langchain.chat_models import ChatAnthropic, ChatOpenAI  # GPT-4 fails to follow the output langchain requires, avoid using for now
from langchain.agents import initialize_agent, create_sql_agent
from langchain.prompts import PromptTemplate
from langchain.agents.agent_toolkits import SQLDatabaseToolkit
from langchain.chains.conversation.memory import ConversationSummaryBufferMemory

from mindsdb.integrations.handlers.openai_handler.constants import CHAT_MODELS as OPEN_AI_CHAT_MODELS
from mindsdb.integrations.handlers.langchain_handler.mindsdb_database_agent import MindsDBSQL
from mindsdb.integrations.handlers.langchain_handler.tools import setup_tools
from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.integrations.utilities.handler_utils import get_api_key
from mindsdb.utilities import log


_DEFAULT_MODEL = 'gpt-3.5-turbo'
_DEFAULT_MAX_TOKENS = 2048  # requires more than vanilla OpenAI due to ongoing summarization and 3rd party input
_DEFAULT_AGENT_MODEL = 'zero-shot-react-description'
_DEFAULT_AGENT_TOOLS = ['python_repl', 'wikipedia']  # these require no additional arguments
_ANTHROPIC_CHAT_MODELS = {'claude-2', 'claude-instant-1'}
_PARSING_ERROR_PREFIX = 'Could not parse LLM output: `'

class LangChainHandler(BaseMLEngine):
    """
    This is a MindsDB integration for the LangChain library, which provides a unified interface for interacting with
    various large language models (LLMs).

    Currently, this integration supports exposing OpenAI and Anthrophic's LLMs with normal text completion support. They are then
    wrapped in a zero shot react description agent that offers a few third party tools out of the box, with support
    for additional ones if an API key is provided. Ongoing memory is also provided.

    Full tool support list:
        - wikipedia
        - python_repl
        - serper.dev search
    """
    name = 'langchain'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generative = True
        self.stops = []
        self.default_mode = 'default'  # can also be 'conversational' or 'conversational-full'
        self.engine_to_supported_modes = {
            'openai': ['default', 'conversational', 'conversational-full', 'image'],
            'anthropic': ['default', 'conversational', 'conversational-full']
        }
        self.default_model = _DEFAULT_MODEL
        self.default_max_tokens = _DEFAULT_MAX_TOKENS
        self.default_agent_model = _DEFAULT_AGENT_MODEL
        self.default_agent_tools = _DEFAULT_AGENT_TOOLS
        self.write_privileges = False  # if True, this agent is able to write into other active mindsdb integrations

    # TODO (ref #7496): modify handler_utils.get_api_key to check for prefix in all sources, update usage in all handlers, deprecate  # noqa
    def _get_serper_api_key(self, args, strict=True):
        if 'serper_api_key' in args:
            return args['serper_api_key']
        # 2
        connection_args = self.engine_storage.get_connection_args()
        if 'serper_api_key' in connection_args:
            return connection_args['serper_api_key']
        # 3
        api_key = os.getenv('SERPER_API_KEY')  # e.g. "OPENAI_API_KEY"
        if api_key is not None:
            return api_key

        if strict:
            raise Exception(f'Missing API key serper_api_key. Either re-create this ML_ENGINE specifying the `serper_api_key` parameter,\
                 or re-create this model and pass the API key with `USING` syntax.')  # noqa

    def create(self, target, args=None, **kwargs):
        self.write_privileges = args.get('using', {}).get('writer', self.write_privileges)
        self.default_agent_tools = args.get('tools', self.default_agent_tools)

        args = args['using']
        args['target'] = target
        
        available_models = {*OPEN_AI_CHAT_MODELS, *_ANTHROPIC_CHAT_MODELS}
        if not args.get('model_name'):
            args['model_name'] = self.default_model
        elif args['model_name'] not in available_models:
            raise Exception(f"Invalid model name. Please use one of {available_models}")

        if not args.get('mode'):
            args['mode'] = self.default_mode

        supported_modes = self.engine_to_supported_modes['openai']
        if args['model_name'] in _ANTHROPIC_CHAT_MODELS:
            supported_modes = self.engine_to_supported_modes['anthropic']
        if args['mode'] not in supported_modes:
            raise Exception(f"Invalid operation mode. Please use one of {supported_modes}")

        self.model_storage.json_set('args', args)

    @staticmethod
    def create_validation(target, args=None, **kwargs):
        if 'using' not in args:
            raise Exception("LangChain engine requires a USING clause! Refer to its documentation for more details.")
        else:
            args = args['using']

        if args.get('mode') != 'conversational':
            if len(set(args.keys()) & {'prompt_template'}) == 0:
                raise Exception('Please provide a `prompt_template` for this engine.')

    def predict(self, df, args=None):
        """
        Dispatch is performed depending on the underlying model type. Currently, only the default text completion
        is supported.
        """
        executor = args['executor']  # used as tool in custom tool for the agent to have mindsdb-wide access
        pred_args = args['predict_params'] if args else {}
        args = self.model_storage.json_get('args')
        args['executor'] = executor

        df = df.reset_index(drop=True)

        if args.get('mode') != 'conversational':
            if 'prompt_template' not in args and 'prompt_template' not in pred_args:
                raise Exception(f"This model expects a prompt template, please provide one.")

        if 'stops' in pred_args:
            self.stops = pred_args['stops']

        # TODO: offload creation to the `create` method instead for faster inference?

        modal_dispatch = {
            'default': 'default_completion',
            'sql_agent': 'sql_agent_completion',
        }

        agent_creation_method = modal_dispatch.get(args.get('modal_dispatch', 'default'), 'default_completion')

        if args.get('mode') == 'conversational':
            agent_creation_method = 'conversational_completion'

        # input dataframe can be modified
        agent, df = getattr(self, agent_creation_method)(df, args, pred_args)
        return self.run_agent(df, agent, args, pred_args)

    def _get_chat_model_params(self, args, pred_args):
        model_name = args.get('model_name', self.default_model)
        # Params shared by all models.
        temperature = min(1.0, max(0.0, args.get('temperature', 0.0)))
        max_tokens = pred_args.get('max_tokens', args.get('max_tokens', self.default_max_tokens))
        top_p = pred_args.get('top_p', None)
        timeout = pred_args.get('request_timeout', None)
        serper_api_key = self._get_serper_api_key(args, strict=False)
        model_kwargs = {}
        if model_name in _ANTHROPIC_CHAT_MODELS:
            model_kwargs['model'] = model_name
            model_kwargs['temperature'] = temperature
            model_kwargs['max_tokens_to_sample'] = max_tokens
            model_kwargs['top_p'] = top_p
            model_kwargs['timeout'] = timeout
            model_kwargs['stop_sequences'] = pred_args.get('stop_sequences', None)
            model_kwargs['serper_api_key'] = serper_api_key
            model_kwargs['anthropic_api_key'] = get_api_key('anthropic', args, self.engine_storage)
        else:
            # OpenAI
            model_kwargs['model_name'] = model_name
            model_kwargs['temperature'] = temperature
            model_kwargs['max_tokens'] = max_tokens
            model_kwargs['top_p'] = top_p
            model_kwargs['request_timeout'] = timeout
            model_kwargs['serper_api_key'] = serper_api_key
            model_kwargs['frequency_penalty'] = pred_args.get('frequency_penalty', None)
            model_kwargs['presence_penalty'] = pred_args.get('presence_penalty', None)
            model_kwargs['n'] = pred_args.get('n', None)
            model_kwargs['best_of'] = pred_args.get('best_of', None)
            model_kwargs['logit_bias'] = pred_args.get('logit_bias', None)
            model_kwargs['openai_api_key'] = get_api_key('openai', args, self.engine_storage)

        model_kwargs = {k: v for k, v in model_kwargs.items() if v is not None}  # filter out None values
        return model_kwargs

    def _create_chat_model(self, args, pred_args):
        model_kwargs = self._get_chat_model_params(args, pred_args)
        model_name = args.get('model_name', self.default_model)

        if model_name in _ANTHROPIC_CHAT_MODELS:
            return ChatAnthropic(**model_kwargs)
        elif model_name in OPEN_AI_CHAT_MODELS:
            return ChatOpenAI(**model_kwargs)
        else:
            return OpenAI(**model_kwargs)

    def conversational_completion(self, df, args=None, pred_args=None):
        pred_args = pred_args if pred_args else {}

        # langchain tool setup
        model_kwargs = self._get_chat_model_params(args, pred_args)
        llm = self._create_chat_model(args, pred_args)
        max_tokens = pred_args.get('max_tokens', args.get('max_tokens', self.default_max_tokens))
        tools = setup_tools(llm,
                            model_kwargs,
                            pred_args,
                            args['executor'],
                            self.default_agent_tools,
                            self.write_privileges)

        memory = ConversationSummaryBufferMemory(llm=llm,
                                                 max_token_limit=max_tokens,
                                                 memory_key="chat_history")

        # fill memory

        # system prompt
        prompt = args['prompt_template']
        if 'prompt_template' in pred_args and pred_args['prompt_template'] is not None:
            prompt = pred_args['prompt_template']
        if 'context' in pred_args:
            prompt += '\n\n' + 'Useful information:\n' + pred_args['context'] + '\n'
        memory.chat_memory.messages.insert(0, SystemMessage(content=prompt))

        # user - assistant conversation. get all except the last message
        for row in df[:-1].to_dict('records'):
            question = row[args['user_column']]
            answer = row[args['assistant_column']]

            if question:
                memory.chat_memory.add_user_message(question)
            if answer:
                memory.chat_memory.add_ai_message(answer)

        # use last message as prompt, remove other questions
        df.iloc[:-1, df.columns.get_loc(args['user_column'])] = ''

        agent_name = AgentType.CONVERSATIONAL_REACT_DESCRIPTION
        agent = initialize_agent(
            tools,
            llm,
            memory=memory,
            agent=agent_name,
            max_iterations=pred_args.get('max_iterations', 3),
            verbose=pred_args.get('verbose', args.get('verbose', False)),
            handle_parsing_errors=False,
        )

        # setup model description
        description = {
            'allowed_tools': [agent.agent.allowed_tools],  # packed as list to avoid additional rows
            'agent_type': agent_name,
            'max_iterations': agent.max_iterations,
            'memory_type': memory.__class__.__name__,
        }

        description = {**description, **model_kwargs}
        description.pop('openai_api_key', None)
        self.model_storage.json_set('description', description)

        return agent, df

    def default_completion(self, df, args=None, pred_args=None):
        """
        Mostly follows the logic of the OpenAI handler, but with a few additions:
            - setup the langchain toolkit
            - setup the langchain agent (memory included)
            - setup information to be published when describing the model

        Ref link from the LangChain documentation on how to accomplish the first two items:
            - python.langchain.com/en/latest/modules/agents/agents/custom_agent.html
        """
        pred_args = pred_args if pred_args else {}
        pred_args['tools'] = args.get('tools') if 'tools' not in pred_args else pred_args.get('tools', [])

        # langchain tool setup
        llm = self._create_chat_model(args, pred_args)
        max_tokens = pred_args.get('max_tokens', args.get('max_tokens', self.default_max_tokens))
        model_kwargs = self._get_chat_model_params(args, pred_args)
        tools = setup_tools(llm,
                            model_kwargs,
                            pred_args,
                            args['executor'],
                            self.default_agent_tools,
                            self.write_privileges)

        # langchain agent setup
        memory = ConversationSummaryBufferMemory(llm=llm, max_token_limit=max_tokens)
        agent_name = pred_args.get('agent_name', args.get('agent_name', self.default_agent_model))
        agent = initialize_agent(
            tools,
            llm,
            memory=memory,
            agent=agent_name,
            max_iterations=pred_args.get('max_iterations', 3),
            verbose=pred_args.get('verbose', args.get('verbose', False)),
            handle_parsing_errors=False,
        )

        # setup model description
        description = {
            'allowed_tools': [agent.agent.allowed_tools],   # packed as list to avoid additional rows
            'agent_type': agent_name,
            'max_iterations': agent.max_iterations,
            'memory_type': memory.__class__.__name__,
        }
        description = {**description, **model_kwargs}
        description.pop('openai_api_key', None)
        self.model_storage.json_set('description', description)

        return agent, df

    def run_agent(self, df, agent, args, pred_args):
        # TODO abstract prompt templating into a common utility method, this is also used in vanilla OpenAI
        if 'prompt_template' in pred_args:
            base_template = pred_args['prompt_template']   # override with predict-time template if available
        elif 'prompt_template' in args:
            base_template = args['prompt_template']  # use create-time template if not
        else:
            base_template = '{{question}}'  # default template otherwise

        input_variables = []
        matches = list(re.finditer("{{(.*?)}}", base_template))

        for m in matches:
            input_variables.append(m[0].replace('{', '').replace('}', ''))

        empty_prompt_ids = np.where(df[input_variables].isna().all(axis=1).values)[0]

        base_template = base_template.replace('{{', '{').replace('}}', '}')
        prompts = []

        for i, row in df.iterrows():
            if i not in empty_prompt_ids:
                prompt = PromptTemplate(input_variables=input_variables, template=base_template)
                kwargs = {}
                for col in input_variables:
                    kwargs[col] = row[col] if row[col] is not None else ''  # add empty quote if data is missing
                prompts.append(prompt.format(**kwargs))
            elif row.get(args['user_column']):
                # just add prompt
                prompts.append(row[args['user_column']])

        def _completion(agent, prompts):
            # TODO: ensure that agent completion plus prompt match the maximum allowed by the user
            # TODO: use async API if possible for parallelized completion
            completions = []
            for prompt in prompts:
                if not prompt:
                    # skip empty values
                    completions.append('')
                    continue
                try:
                    completions.append(agent.run(prompt))
                except ValueError as e:
                    # Handle parsing errors ourselves instead of using handle_parsing_errors=True in initialize_agent.
                    response = str(e)
                    if not response.startswith(_PARSING_ERROR_PREFIX):
                        completions.append(f'agent failed with error:\n{str(e)}...')
                    else:
                        # By far the most common error is a Langchain parsing error. Some OpenAI models
                        # always output a response formatted correctly. Anthropic, and others, sometimes just output
                        # the answer as text (when not using tools), even when prompted to output in a specific format.
                        # As a somewhat dirty workaround, we accept the output formatted incorrectly and use it as a response.
                        #
                        # Ideally, in the future, we would write a parser that is more robust and flexible than the one Langchain uses.
                        response = response.lstrip(_PARSING_ERROR_PREFIX).rstrip('`')
                        log.logger.info(f"Agent failure, salvaging response...")
                        completions.append(response)
                except Exception as e:
                    completions.append(f'agent failed with error:\n{str(e)}...')
            return [c for c in completions]

        completion = _completion(agent, prompts)

        # add null completion for empty prompts
        for i in sorted(empty_prompt_ids):
            completion.insert(i, None)

        pred_df = pd.DataFrame(completion, columns=[args['target']])

        return pred_df

    def describe(self, attribute: Optional[str] = None) -> pd.DataFrame:
        info = self.model_storage.json_get('description')

        if attribute == 'info':
            if info is None:
                # we do this due to the huge amount of params that can be changed
                #  at prediction time to customize behavior.
                # for them, we report the last observed value
                raise Exception('This model needs to be used before it can be described.')

            description = pd.DataFrame(info)
            return description
        else:
            tables = ['info']
            return pd.DataFrame(tables, columns=['tables'])

    def finetune(self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        raise NotImplementedError('Fine-tuning is not supported for LangChain models')

    def sql_agent_completion(self, df, args=None, pred_args=None):
        """This completion will be used to answer based on information passed by any MindsDB DB or API engine."""
        db = MindsDBSQL(engine=args['executor'], metadata=args['executor'].session.integration_controller)
        toolkit = SQLDatabaseToolkit(db=db)
        model_name = args.get('model_name', self.default_model)
        llm = OpenAI(temperature=0) if model_name not in OPEN_AI_CHAT_MODELS else ChatOpenAI(temperature=0)
        agent = create_sql_agent(
            llm=llm,
            toolkit=toolkit,
            verbose=True
        )
        return agent, df



import copy
import os
import sys
import json
import base64
import datetime as dt
from copy import deepcopy
from multiprocessing.pool import ThreadPool

import pandas as pd
from dateutil.parser import parse as parse_datetime

from sqlalchemy import func, null
import numpy as np

import mindsdb.interfaces.storage.db as db
from mindsdb.interfaces.storage.fs import FsStore
from mindsdb.utilities.config import Config
from mindsdb.utilities.json_encoder import json_serialiser
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE
from mindsdb.interfaces.model.functions import (
    get_model_record,
    get_model_records
)
from mindsdb.interfaces.storage.json import get_json_storage
from mindsdb.interfaces.storage.model_fs import ModelStorage
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.functions import resolve_model_identifier
import mindsdb.utilities.profiler as profiler

IS_PY36 = sys.version_info[1] <= 6


def delete_model_storage(model_id, ctx_dump):
    try:
        ctx.load(ctx_dump)
        modelStorage = ModelStorage(model_id)
        modelStorage.delete()
    except Exception as e:
        print(f'Something went wrong during deleting storage of model {model_id}: {e}')


class ModelController():
    config: Config

    def __init__(self) -> None:
        self.config = Config()

    def get_model_data(self, name: str = None, predictor_record=None, ml_handler_name='lightwood') -> dict:
        if predictor_record is None:
            predictor_record = get_model_record(except_absent=True, name=name, ml_handler_name=ml_handler_name)

        data = deepcopy(predictor_record.data)
        data['dtype_dict'] = predictor_record.dtype_dict
        data['created_at'] = str(parse_datetime(str(predictor_record.created_at).split('.')[0]))
        data['updated_at'] = str(parse_datetime(str(predictor_record.updated_at).split('.')[0]))
        data['training_start_at'] = predictor_record.training_start_at
        data['training_stop_at'] = predictor_record.training_stop_at
        data['predict'] = predictor_record.to_predict[0]
        data['update'] = predictor_record.update_status
        data['mindsdb_version'] = predictor_record.mindsdb_version
        data['name'] = predictor_record.name
        data['code'] = predictor_record.code
        data['problem_definition'] = predictor_record.learn_args
        data['fetch_data_query'] = predictor_record.fetch_data_query
        data['active'] = predictor_record.active
        data['status'] = predictor_record.status
        data['id'] = predictor_record.id
        data['version'] = predictor_record.version

        json_storage = get_json_storage(
            resource_id=predictor_record.id
        )
        data['json_ai'] = json_storage.get('json_ai')

        if data.get('accuracies', None) is not None:
            if len(data['accuracies']) > 0:
                data['accuracy'] = float(np.mean(list(data['accuracies'].values())))
        return data

    def get_reduced_model_data(self, name: str = None, predictor_record=None, ml_handler_name='lightwood') -> dict:
        full_model_data = self.get_model_data(name=name, predictor_record=predictor_record, ml_handler_name=ml_handler_name)
        reduced_model_data = {}
        for k in ['id', 'name', 'version', 'is_active', 'predict', 'status',
                  'problem_definition', 'current_phase', 'accuracy', 'data_source', 'update', 'active',
                  'mindsdb_version', 'error', 'created_at', 'fetch_data_query']:
            reduced_model_data[k] = full_model_data.get(k, None)

        reduced_model_data['training_time'] = None
        if full_model_data.get('training_start_at') is not None:
            if full_model_data.get('training_stop_at') is not None:
                reduced_model_data['training_time'] = (
                    full_model_data.get('training_stop_at')
                    - full_model_data.get('training_start_at')
                )
            elif full_model_data.get('status') == 'training':
                reduced_model_data['training_time'] = (
                    dt.datetime.now()
                    - full_model_data.get('training_start_at')
                )
            if reduced_model_data['training_time'] is not None:
                reduced_model_data['training_time'] = (
                    reduced_model_data['training_time']
                    - dt.timedelta(microseconds=reduced_model_data['training_time'].microseconds)
                )

        return reduced_model_data

    def describe_model(self, session, project_name, model_name, attribute, version=None):
        model_record = get_model_record(
            name=model_name,
            version=version,
            project_name=project_name,
            except_absent=True
        )
        integration_record = db.Integration.query.get(model_record.integration_id)

        ml_handler_base = session.integration_controller.get_handler(integration_record.name)

        ml_handler = ml_handler_base.get_ml_handler(model_record.id)

        if attribute is None:
            model_info = self.get_model_info(model_record)

            try:
                df = ml_handler.describe(attribute)
            except NotImplementedError:
                df = pd.DataFrame()

            # expecting list of attributes in first column df
            attributes = []
            if len(df) > 0 and len(df.columns) > 0:
                attributes = list(df[df.columns[0]])
                if len(attributes) == 1 and isinstance(attributes[0], list):
                    # first cell already has a list
                    attributes = attributes[0]

            model_info.insert(0, 'TABLES', [attributes])
            return model_info
        else:
            return ml_handler.describe(attribute)

    def get_model(self, name, version=None, ml_handler_name=None, project_name=None):
        show_active = True if version is None else None
        model_record = get_model_record(
            active=show_active,
            version=version,
            name=name,
            ml_handler_name=ml_handler_name,
            project_name=project_name)
        return self.get_reduced_model_data(predictor_record=model_record)

    def get_models(self, with_versions=False, ml_handler_name=None, integration_id=None,
                   project_name=None):
        models = []
        show_active = True if with_versions is False else None
        for model_record in get_model_records(active=show_active, ml_handler_name=ml_handler_name,
                                              integration_id=integration_id, project_name=project_name):
            model_data = self.get_reduced_model_data(predictor_record=model_record)
            models.append(model_data)
        return models

    def delete_model(self, model_name: str, project_name: str = 'mindsdb', version=None):
        from mindsdb.interfaces.database.database import DatabaseController

        project_record = db.Project.query.filter(
            (func.lower(db.Project.name) == func.lower(project_name))
            & (db.Project.company_id == ctx.company_id)
            & (db.Project.deleted_at == null())
        ).first()
        if project_record is None:
            raise Exception(f"Project '{project_name}' does not exists")

        database_controller = DatabaseController()

        project = database_controller.get_project(project_name)

        if version is None:
            # Delete latest version
            predictors_records = get_model_records(
                name=model_name,
                project_id=project.id,
                active=None,
            )
        else:
            predictors_records = get_model_records(
                name=model_name,
                project_id=project.id,
                version=version,
            )
        if len(predictors_records) == 0:
            raise Exception(f"Model '{model_name}' does not exist")

        is_cloud = self.config.get('cloud', False)
        if is_cloud:
            for predictor_record in predictors_records:
                model_data = self.get_model_data(predictor_record=predictor_record)
                if (
                    model_data.get('status') in ['generating', 'training']
                    and isinstance(model_data.get('created_at'), str) is True
                    and (dt.datetime.now() - parse_datetime(model_data.get('created_at'))) < dt.timedelta(hours=1)
                ):
                    raise Exception('You are unable to delete models currently in progress, please wait before trying again')

        for predictor_record in predictors_records:
            if is_cloud:
                predictor_record.deleted_at = dt.datetime.now()
            else:
                db.session.delete(predictor_record)
        db.session.commit()

        # region delete storages
        if len(predictors_records) > 1:
            ctx_dump = ctx.dump()
            with ThreadPool(min(len(predictors_records), 100)) as pool:
                pool.starmap(delete_model_storage, [(record.id, ctx_dump) for record in predictors_records])
        else:
            modelStorage = ModelStorage(predictors_records[0].id)
            modelStorage.delete()
        # endregion

    def rename_model(self, old_name, new_name):
        model_record = get_model_record(name=new_name)
        if model_record is None:
            raise Exception(f"Model with name '{new_name}' already exists")

        for model_record in get_model_records(name=old_name):
            model_record.name = new_name
        db.session.commit()

    @staticmethod
    def _get_data_integration_ref(statement, database_controller):
        # TODO use database_controller handler_controller internally
        data_integration_ref = None
        fetch_data_query = None
        if statement.integration_name is not None:
            fetch_data_query = statement.query_str
            integration_name = statement.integration_name.parts[0]

            databases_meta = database_controller.get_dict()
            data_integration_meta = databases_meta[integration_name]
            # TODO improve here. Suppose that it is view
            if data_integration_meta['type'] == 'project':
                data_integration_ref = {
                    'type': 'view'
                }
            else:
                data_integration_ref = {
                    'type': 'integration',
                    'id': data_integration_meta['id']
                }
        return data_integration_ref, fetch_data_query

    def prepare_create_statement(self, statement, database_controller):
        # extract data from Create model or Retrain statement and prepare it for using in crate and retrain functions
        project_name = statement.name.parts[0].lower()
        model_name = statement.name.parts[1].lower()

        sql_task = None
        if statement.task is not None:
            sql_task = statement.task.to_string()
        problem_definition = {
            '__mdb_sql_task': sql_task
        }
        if statement.targets is not None:
            problem_definition['target'] = statement.targets[0].parts[-1]

        data_integration_ref, fetch_data_query = self._get_data_integration_ref(statement, database_controller)

        label = None
        if statement.using is not None:
            label = statement.using.pop('tag', None)

            problem_definition['using'] = statement.using

        if statement.order_by is not None:
            problem_definition['timeseries_settings'] = {
                'is_timeseries': True,
                'order_by': getattr(statement, 'order_by')[0].field.parts[-1]
            }
            for attr in ['horizon', 'window']:
                if getattr(statement, attr) is not None:
                    problem_definition['timeseries_settings'][attr] = getattr(statement, attr)

            if statement.group_by is not None:
                problem_definition['timeseries_settings']['group_by'] = [col.parts[-1] for col in statement.group_by]

        join_learn_process = False
        if 'join_learn_process' in problem_definition.get('using', {}):
            join_learn_process = problem_definition['using']['join_learn_process']
            del problem_definition['using']['join_learn_process']

        return dict(
            model_name=model_name,
            project_name=project_name,
            data_integration_ref=data_integration_ref,
            fetch_data_query=fetch_data_query,
            problem_definition=problem_definition,
            join_learn_process=join_learn_process,
            label=label
        )

    def create_model(self, statement, ml_handler):
        params = self.prepare_create_statement(statement, ml_handler.database_controller)

        existing_projects_meta = ml_handler.database_controller.get_dict(filter_type='project')
        if params['project_name'] not in existing_projects_meta:
            raise Exception(f"Project '{params['project_name']}' does not exist.")

        project = ml_handler.database_controller.get_project(name=params['project_name'])
        project_tables = project.get_tables()
        if params['model_name'] in project_tables:
            raise Exception(f"Error: model '{params['model_name']}' already exists in project {params['project_name']}!")

        predictor_record = ml_handler.learn(**params)

        return self.get_model_info(predictor_record)

    def retrain_model(self, statement, ml_handler):
        # active setting
        set_active = True
        if statement.using is not None:
            set_active = statement.using.pop('active', True)
            if set_active in ('0', 0, None):
                set_active = False

        params = self.prepare_create_statement(statement, ml_handler.database_controller)

        base_predictor_record = get_model_record(
            name=params['model_name'],
            project_name=params['project_name'],
            active=True
        )

        model_name = params['model_name']
        if base_predictor_record is None:
            raise Exception(f"Error: model '{model_name}' does not exist")

        if params['data_integration_ref'] is None:
            params['data_integration_ref'] = base_predictor_record.data_integration_ref
        if params['fetch_data_query'] is None:
            params['fetch_data_query'] = base_predictor_record.fetch_data_query

        problem_definition = base_predictor_record.learn_args.copy()
        problem_definition.update(params['problem_definition'])
        params['problem_definition'] = problem_definition

        params['is_retrain'] = True
        params['set_active'] = set_active
        predictor_record = ml_handler.learn(**params)

        return self.get_model_info(predictor_record)

    def prepare_finetune_statement(self, statement, database_controller):
        project_name, model_name, model_version = resolve_model_identifier(statement.name)
        if project_name is None:
            project_name = 'mindsdb'
        data_integration_ref, fetch_data_query = self._get_data_integration_ref(statement, database_controller)

        set_active = True
        if statement.using is not None:
            set_active = statement.using.pop('active', True)
            if set_active in ('0', 0, None):
                set_active = False

        label = None
        args = {}
        if statement.using is not None:
            label = statement.using.pop('tag', None)
            args = statement.using

        join_learn_process = args.pop('join_learn_process', False)

        base_predictor_record = get_model_record(
            name=model_name,
            project_name=project_name,
            version=model_version,
            active=True if model_version is None else None
        )

        if data_integration_ref is None:
            data_integration_ref = base_predictor_record.data_integration_ref
        if fetch_data_query is None:
            fetch_data_query = base_predictor_record.fetch_data_query

        return dict(
            model_name=model_name,
            project_name=project_name,
            data_integration_ref=data_integration_ref,
            fetch_data_query=fetch_data_query,
            base_model_version=model_version,
            args=args,
            join_learn_process=join_learn_process,
            label=label,
            set_active=set_active
        )

    @profiler.profile()
    def finetune_model(self, statement, ml_handler):
        params = self.prepare_finetune_statement(statement, ml_handler.database_controller)
        predictor_record = ml_handler.finetune(**params)
        return self.get_model_info(predictor_record)

    def update_model(self, session, project_name: str, model_name: str, problem_definition, version=None):

        model_record = get_model_record(
            name=model_name,
            version=version,
            project_name=project_name,
            except_absent=True
        )
        integration_record = db.Integration.query.get(model_record.integration_id)

        ml_handler_base = session.integration_controller.get_handler(integration_record.name)

        ml_handler = ml_handler_base.get_ml_handler(model_record.id)
        if not hasattr(ml_handler, 'update'):
            raise Exception("ML handler doesn't updating")

        ml_handler.update(args=problem_definition)

        # update model record
        if 'using' in problem_definition:
            learn_args = copy.deepcopy(model_record.learn_args)
            learn_args['using'].update(problem_definition['using'])
            model_record.learn_args = learn_args
            db.session.commit()

    def get_model_info(self, predictor_record):
        from mindsdb.interfaces.database.projects import ProjectController
        projects_controller = ProjectController()
        project = projects_controller.get(id=predictor_record.project_id)

        columns = ['NAME', 'ENGINE', 'PROJECT', 'ACTIVE', 'VERSION', 'STATUS', 'ACCURACY', 'PREDICT', 'UPDATE_STATUS',
                   'MINDSDB_VERSION', 'ERROR', 'SELECT_DATA_QUERY', 'TRAINING_OPTIONS', 'TAG']

        project_name = project.name
        model = project.get_model_by_id(model_id=predictor_record.id)
        table_name = model['name']
        table_meta = model['metadata']
        record = [
            table_name, table_meta['engine'], project_name, table_meta['active'], table_meta['version'], table_meta['status'],
            table_meta['accuracy'], table_meta['predict'], table_meta['update_status'],
            table_meta['mindsdb_version'], table_meta['error'], table_meta['select_data_query'],
            str(table_meta['training_options']), table_meta['label']
        ]

        return pd.DataFrame([record], columns=columns)

    def update_model_version(self, models, active=None):
        if active is None:
            raise NotImplementedError(f'Update is not supported')

        if active in ('0', 0, False):
            active = False
        else:
            active = True

        if active is False:
            raise NotImplementedError('Only setting active version is possible')

        if len(models) != 1:
            raise Exception('Only one version can be updated')

        # update
        model = models[0]

        model_record = get_model_record(
            name=model['NAME'],
            project_name=model['PROJECT'],
            version=model['VERSION'],
            active=None
        )

        model_record.active = True

        # deactivate current active version
        model_records = db.Predictor.query.filter(
            db.Predictor.name == model_record.name,
            db.Predictor.project_id == model_record.project_id,
            db.Predictor.active == True,
            db.Predictor.company_id == ctx.company_id,
            db.Predictor.id != model_record.id
        )
        for p in model_records:
            p.active = False

        db.session.commit()

    def delete_model_version(self, models):
        if len(models) == 0:
            raise Exception(f"Version to delete is not found")

        for model in models:
            model_record = get_model_record(
                name=model['NAME'],
                project_name=model['PROJECT'],
                version=model['VERSION'],
                active=None
            )
            if model_record.active:
                raise Exception(f"Can't remove active version: {model['PROJECT']}.{model['NAME']}.{model['VERSION']}")

            is_cloud = self.config.get('cloud', False)
            if is_cloud:
                model_record.deleted_at = dt.datetime.now()
            else:
                db.session.delete(model_record)
            modelStorage = ModelStorage(model_record.id)
            modelStorage.delete()

        db.session.commit()



from typing import List

from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb.utilities.log import get_log

from mindsdb_sql.parser import ast
from mindsdb.integrations.handlers.utilities.query_utilities import (
    SELECTQueryParser,
    SELECTQueryExecutor,
)

import pandas as pd
import re
from youtube_transcript_api import YouTubeTranscriptApi
from youtube_transcript_api.formatters import JSONFormatter

logger = get_log("integrations.youtube_handler")


class YoutubeCommentsTable(APITable):
    """Youtube List Comments  by video id Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the youtube  "commentThreads()" API endpoint
        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query
        Returns
        -------
        pd.DataFrame
            youtube "commentThreads()" matching the query
        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        conditions = extract_comparison_conditions(query.where)

        order_by_conditions = {}
        clubs_kwargs = {}

        if query.order_by and len(query.order_by) > 0:
            order_by_conditions["columns"] = []
            order_by_conditions["ascending"] = []

            for an_order in query.order_by:
                if an_order.field.parts[0] != "id":
                    next
                if an_order.field.parts[1] in self.get_columns():
                    order_by_conditions["columns"].append(an_order.field.parts[1])

                    if an_order.direction == "ASC":
                        order_by_conditions["ascending"].append(True)
                    else:
                        order_by_conditions["ascending"].append(False)
                else:
                    raise ValueError(
                        f"Order by unknown column {an_order.field.parts[1]}"
                    )

        for a_where in conditions:
            if a_where[1] == "video_id":
                if a_where[0] != "=":
                    raise ValueError("Unsupported where operation for youtube video id")
                clubs_kwargs["type"] = a_where[2]
            else:
                raise ValueError(f"Unsupported where argument {a_where[1]}")

        youtube_comments_df = self.call_youtube_comments_api(a_where[2])

        selected_columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")

        if len(youtube_comments_df) == 0:
            youtube_comments_df = pd.DataFrame([], columns=selected_columns)
        else:
            youtube_comments_df.columns = self.get_columns()
            for col in set(youtube_comments_df.columns).difference(
                set(selected_columns)
            ):
                youtube_comments_df = youtube_comments_df.drop(col, axis=1)

            if len(order_by_conditions.get("columns", [])) > 0:
                youtube_comments_df = youtube_comments_df.sort_values(
                    by=order_by_conditions["columns"],
                    ascending=order_by_conditions["ascending"],
                )

        if query.limit:
            youtube_comments_df = youtube_comments_df.head(query.limit.value)

        return youtube_comments_df

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses
        Returns
        -------
        List[str]
            List of columns
        """
        return ["user_id", "display_name", "comment"]

    def call_youtube_comments_api(self, video_id):
        """Pulls all the records from the given youtube api end point and returns it select()

        Returns
        -------
        pd.DataFrame of all the records of the "commentThreads()" API end point
        """

        resource = (
            self.handler.connect()
            .commentThreads()
            .list(part="snippet", videoId=video_id, textFormat="plainText")
        )

        video_cols = self.get_columns()
        all_youtube_comments_df = pd.DataFrame(columns=video_cols)
        data = []

        while resource:
            comments = resource.execute()
            for comment in comments["items"]:
                user_id = comment["snippet"]["topLevelComment"]["snippet"][
                    "authorChannelId"
                ]["value"]
                display_name = comment["snippet"]["topLevelComment"]["snippet"][
                    "authorDisplayName"
                ]
                comment_text = comment["snippet"]["topLevelComment"]["snippet"][
                    "textDisplay"
                ]
                data = pd.DataFrame(
                    {
                        "user_id": [user_id],
                        "display_name": [display_name],
                        "comment": [comment_text],
                    }
                )
                all_youtube_comments_df = pd.concat(
                    [all_youtube_comments_df, data], ignore_index=True
                )
            if "nextPageToken" in comments:
                resource = (
                    self.handler.connect()
                    .commentThreads()
                    .list(
                        part="snippet",
                        videoId=video_id,
                        textFormat="plainText",
                        pageToken=comments["nextPageToken"],
                    )
                )
            else:
                break

        return all_youtube_comments_df


class YoutubeChannelsTable(APITable):

    """Youtube Channel Info  by channel id Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        select_statement_parser = SELECTQueryParser(
            query, "channel", self.get_columns()
        )

        (
            selected_columns,
            where_conditions,
            order_by_conditions,
            result_limit,
        ) = select_statement_parser.parse_query()

        channel_id = None
        for op, arg1, arg2 in where_conditions:
            if arg1 == "channel_id":
                if op == "=":
                    channel_id = arg2
                    break
                else:
                    raise NotImplementedError(
                        "Only '=' operator is supported for channel_id column."
                    )

        if not channel_id:
            raise NotImplementedError("channel_id has to be present in where clause.")

        channel_df = self.get_channel_details(channel_id)

        select_statement_executor = SELECTQueryExecutor(
            channel_df, selected_columns, where_conditions, order_by_conditions
        )

        channel_df = select_statement_executor.execute_query()

        return channel_df

    def get_channel_details(self, channel_id):
        details = (
            self.handler.connect()
            .channels()
            .list(part="statistics,snippet,contentDetails", id=channel_id)
            .execute()
        )
        snippet = details["items"][0]["snippet"]
        statistics = details["items"][0]["statistics"]
        data = {
            "country": snippet["country"],
            "description": snippet["description"],
            "creation_date": snippet["publishedAt"],
            "title": snippet["title"],
            "subscriber_count": statistics["subscriberCount"],
            "video_count": statistics["videoCount"],
            "view_count": statistics["viewCount"],
            "channel_id": channel_id,
        }
        return pd.json_normalize(data)

    def get_columns(self) -> List[str]:
        return [
            "country",
            "description",
            "creation_date",
            "title",
            "subscriber_count",
            "video_count",
            "view_count",
            "channel_id",
        ]


class YoutubeVideosTable(APITable):

    """Youtube Video info  by video id Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        select_statement_parser = SELECTQueryParser(query, "video", self.get_columns())

        (
            selected_columns,
            where_conditions,
            order_by_conditions,
            result_limit,
        ) = select_statement_parser.parse_query()

        video_id = None
        for op, arg1, arg2 in where_conditions:
            if arg1 == "video_id":
                if op == "=":
                    video_id = arg2
                    break
                else:
                    raise NotImplementedError(
                        "Only '=' operator is supported for video_id column."
                    )

        if not video_id:
            raise NotImplementedError("video_id has to be present in where clause.")

        video_df = self.get_video_details(video_id)

        select_statement_executor = SELECTQueryExecutor(
            video_df, selected_columns, where_conditions, order_by_conditions
        )

        video_df = select_statement_executor.execute_query()

        return video_df

    def get_video_details(self, video_id):
        details = (
            self.handler.connect()
            .videos()
            .list(part="statistics,snippet,contentDetails", id=video_id)
            .execute()
        )
        items = details.get("items")[0]
        snippet = items["snippet"]
        statistics = items["statistics"]
        content_details = items["contentDetails"]
        transcript = self.get_captions(video_id)
        data = {
            "channel_title": snippet["channelTitle"],
            "comment_count": statistics["commentCount"],
            "description": snippet["description"],
            "like_count": statistics["likeCount"],
            "publish_time": snippet["publishedAt"],
            "title": snippet["title"],
            "transcript": transcript,
            "video_id": video_id,
            "view_count": statistics["viewCount"],
        }
        duration = content_details["duration"]
        parsed_duration = re.search(f"PT(\d+H)?(\d+M)?(\d+S)", duration).groups()
        duration_str = ""
        for d in parsed_duration:
            if d:
                duration_str += f"{d[:-1]}:"
        data["duration_str"] = duration_str.strip(":")
        return pd.json_normalize(data)

    def get_captions(self, video_id):
        try:
            transcript_response = YouTubeTranscriptApi.get_transcript(
                video_id, preserve_formatting=True
            )
            json_formatted_transcript = JSONFormatter().format_transcript(
                transcript_response, indent=2
            )
            return json_formatted_transcript

        except Exception as e:
            logger.error(
                f"Encountered an error while fetching transcripts for video ${video_id}: ${e}"
            ),
            return "Transcript not available for this video"

    def get_columns(self) -> List[str]:
        return [
            "channel_title",
            "title",
            "description",
            "publish_time",
            "comment_count",
            "like_count",
            "view_count",
            "view_count",
            "video_id",
            "duration_str",
            "transcript",
        ]



import unittest
from mindsdb.integrations.handlers.strapi_handler.strapi_handler import StrapiHandler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE


class StrapiHandlerTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        connection_data = {
            'host': 'localhost',
            'port': '1337',
            'api_token': 'c56c000d867e95848c',
            'pluralApiIds': ['products', 'sellers']}
        cls.handler = StrapiHandler(name='myshop', connection_data=connection_data)

    def test_0_check_connection(self):
        # Ensure the connection is successful
        self.assertTrue(self.handler.check_connection())

    def test_1_get_table(self):
        assert self.handler.get_tables() is not RESPONSE_TYPE.ERROR

    def test_2_get_columns(self):
        assert self.handler.get_columns('products') is not RESPONSE_TYPE.ERROR

    def test_3_get_data(self):
        # Ensure that you can retrieve data from a table
        data = self.handler.native_query('SELECT * FROM products')
        assert data.type is not RESPONSE_TYPE.ERROR

    def test_4_get_data_with_condition(self):
        # Ensure that you can retrieve data with a condition
        data = self.handler.native_query('SELECT * FROM products WHERE id = 1')
        assert data.type is not RESPONSE_TYPE.ERROR

    def test_5_insert_data(self):
        # Ensure that data insertion is successful
        query = "INSERT INTO myshop.sellers (name, email, sellerid) VALUES ('Ram', 'ram@gmail.com', 'ramu4')"
        result = self.handler.native_query(query)
        self.assertTrue(result)

    def test_6_update_data(self):
        # Ensure that data updating is successful
        query = "UPDATE products SET name = 'test2' WHERE id = 1"
        result = self.handler.native_query(query)
        self.assertTrue(result)


if __name__ == '__main__':
    unittest.main()



"""predictor_version

Revision ID: 1e60096fc817
Revises: 43c52d23845a
Create Date: 2022-11-07 16:43:47.301692

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import text


# revision identifiers, used by Alembic.
revision = '1e60096fc817'
down_revision = '43c52d23845a'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.add_column(sa.Column('label', sa.String(), nullable=True))
        batch_op.add_column(sa.Column('version', sa.Integer(), nullable=True))

    # update current predictor versions
    conn = op.get_bind()
    session = sa.orm.Session(bind=conn)

    key0 = (None, None, None)
    models = conn.execute(sa.text('''
        select company_id, project_id, name, id
        from predictor
        order by company_id, project_id, name, created_at
    ''')).fetchall()
    for model in models:
        key = (model['company_id'], model['project_id'], model['name'].lower())

        # it is different name or project or company
        if key != key0:
            version = 1
            key0 = key

        conn.execute(
            text("""
                update predictor
                set version = :version
                where id = :id
            """), {
                'version': version,
                'id': model['id']
            }
        )
        version += 1

    session.commit()
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.drop_column('version')
        batch_op.drop_column('label')

    # ### end Alembic commands ###



class HANDLER_TYPE:
    __slots__ = ()
    DATA = 'data'
    ML = 'ml'


HANDLER_TYPE = HANDLER_TYPE()


class HANDLER_CONNECTION_ARG_TYPE:
    __slots__ = ()
    STR = 'str'
    INT = 'int'
    BOOL = 'bool'
    URL = 'url'
    PATH = 'path'
    DICT = 'dict'
    PWD = 'pwd'


HANDLER_CONNECTION_ARG_TYPE = HANDLER_CONNECTION_ARG_TYPE()

from mindsdb.interfaces.storage.db import PREDICTOR_STATUS  # noqa



import time
import pickle
import socket
import threading

from walrus import Database
from redis.exceptions import ConnectionError as RedisConnectionError

from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.ml_task_queue.const import ML_TASK_STATUS


def to_bytes(obj: object) -> bytes:
    """ dump object into bytes

        Args:
            obj (object): object to convert

        Returns:
            bytes
    """
    return pickle.dumps(obj, protocol=5)


def from_bytes(b: bytes) -> object:
    """ load object from bytes

        Args:
            b (bytes):

        Returns:
            object
    """
    return pickle.loads(b)


def wait_redis_ping(db: Database, timeout: int = 30):
    """ Wait when redis.ping return True

        Args:
            db (Database): redis db object
            timeout (int): seconds to wait for success ping

        Raises:
            RedisConnectionError: if `ping` did not return `True` within `timeout` seconds
    """
    end_time = time.time() + timeout
    while time.time() <= end_time:
        try:
            if db.ping() is True:
                break
        except RedisConnectionError:
            pass
        time.sleep(2)
    else:
        raise RedisConnectionError


class RedisKey:
    """ The class responsible for unique task keys in redis

        Attributes:
            _base_key (bytes): prefix for keys
    """

    @staticmethod
    def new() -> 'RedisKey':
        timestamp = str(time.time()).replace('.', '')
        return RedisKey(f"{timestamp}-{ctx.company_id}-{socket.gethostname()}".encode())

    def __init__(self, base_key: bytes) -> None:
        self._base_key = base_key

    @property
    def base(self) -> bytes:
        return self._base_key

    @property
    def status(self) -> str:
        return (self._base_key + b'-status').decode()

    @property
    def dataframe(self) -> str:
        return (self._base_key + b'-dataframe').decode()

    @property
    def exception(self) -> str:
        return (self._base_key + b'-exception').decode()


class StatusNotifier(threading.Thread):
    """ Worker that updates task status in redis with fixed frequency
    """

    def __init__(self, redis_key: RedisKey, ml_task_status: ML_TASK_STATUS, db, cache) -> None:
        threading.Thread.__init__(self)
        self.redis_key = redis_key
        self.ml_task_status = ml_task_status
        self.db = db
        self.cache = cache
        self._stop_event = threading.Event()

    def set_status(self, ml_task_status: ML_TASK_STATUS):
        """ change status

            Args:
                ml_task_status (ML_TASK_STATUS): new status
        """
        self.ml_task_status = ml_task_status

    def stop(self) -> None:
        """ stop status updating
        """
        self._stop_event.set()

    def run(self):
        """ start update status with fixed frequency
        """
        while not self._stop_event.is_set():
            wait_redis_ping(self.db)
            self.db.publish(self.redis_key.status, self.ml_task_status.value)
            self.cache.set(self.redis_key.status, self.ml_task_status.value, 180)
            time.sleep(5)



import os
import tempfile
import threading
import time
from pathlib import Path
from typing import Optional

import psutil
from appdirs import user_data_dir


def create_directory(path):
    path = Path(path)
    path.mkdir(mode=0o777, exist_ok=True, parents=True)


def get_root_path():
    mindsdb_path = user_data_dir("mindsdb", "mindsdb")
    return os.path.join(mindsdb_path, "var/")


def get_or_create_data_dir():
    data_dir = user_data_dir("mindsdb", "mindsdb")
    mindsdb_data_dir = os.path.join(data_dir, "var/")

    if os.path.exists(mindsdb_data_dir) is False:
        create_directory(mindsdb_data_dir)

    try:
        assert os.path.exists(mindsdb_data_dir)
        assert os.access(mindsdb_data_dir, os.W_OK) is True
    except Exception:
        raise Exception(
            "MindsDB storage directory does not exist and could not be created"
        )

    return mindsdb_data_dir


def create_dirs_recursive(path):
    if isinstance(path, dict):
        for p in path.values():
            create_dirs_recursive(p)
    elif isinstance(path, str):
        create_directory(path)
    else:
        raise ValueError(f"Wrong path: {path}")


def _get_process_mark_id(unified: bool = False) -> str:
    """Creates a text that can be used to identify process+thread
    Args:
        unified: bool, if True then result will be same for same process+thread
    Returns:
        mark of process+thread
    """
    mark = f"{os.getpid()}-{threading.get_native_id()}"
    if unified is True:
        return mark
    return f"{mark}-{str(time.time()).replace('.', '')}"


def create_process_mark(folder="learn"):
    mark = None
    if os.name == "posix":
        p = Path(tempfile.gettempdir()).joinpath(f"mindsdb/processes/{folder}/")
        p.mkdir(parents=True, exist_ok=True)
        mark = _get_process_mark_id()
        p.joinpath(mark).touch()
    return mark


def set_process_mark(folder: str, mark: str) -> None:
    """touch new file which will be process mark

    Args:
        folder (str): where create the file
        mark (str): file name

    Returns:
        str: process mark
    """
    if os.name != "posix":
        return
    p = Path(tempfile.gettempdir()).joinpath(f"mindsdb/processes/{folder}/")
    p.mkdir(parents=True, exist_ok=True)
    p.joinpath(f"{os.getpid()}-{threading.get_native_id()}-{mark}").touch()
    return mark


def delete_process_mark(folder: str = "learn", mark: Optional[str] = None):
    if mark is None:
        mark = _get_process_mark_id()
    if os.name == "posix":
        p = (
            Path(tempfile.gettempdir())
            .joinpath(f"mindsdb/processes/{folder}/")
            .joinpath(mark)
        )
        if p.exists():
            p.unlink()


def clean_process_marks():
    """delete all existing processes marks"""
    if os.name != "posix":
        return

    p = Path(tempfile.gettempdir()).joinpath("mindsdb/processes/")
    if p.exists() is False:
        return
    for path in p.iterdir():
        if path.is_dir() is False:
            return
        for file in path.iterdir():
            file.unlink()


def clean_unlinked_process_marks():
    """delete marks that does not have corresponded processes/threads"""
    if os.name != "posix":
        return

    p = Path(tempfile.gettempdir()).joinpath("mindsdb/processes/")
    if p.exists() is False:
        return
    for path in p.iterdir():
        if path.is_dir() is False:
            return
        for file in path.iterdir():
            parts = file.name.split("-")
            process_id = int(parts[0])
            thread_id = int(parts[1])

            try:
                process = psutil.Process(process_id)
                if process.status() in (psutil.STATUS_ZOMBIE, psutil.STATUS_DEAD):
                    raise psutil.NoSuchProcess(process_id)

                threads = process.threads()
                try:
                    next(t for t in threads if t.id == thread_id)
                except StopIteration:
                    from mindsdb.utilities.log import get_log

                    get_log(__name__).warning(
                        f"We have mark for process/thread {process_id}/{thread_id} but it does not exists"
                    )
                    file.unlink()

            except psutil.AccessDenied:
                from mindsdb.utilities.log import get_log

                get_log(__name__).warning(f"access to process id: {process_id} denied")

                continue

            except psutil.NoSuchProcess:
                from mindsdb.utilities.log import get_log

                get_log(__name__).warning(
                    f"We have mark for process/thread {process_id}/{thread_id} but it does not exists"
                )
                file.unlink()



import pandas as pd
from mindsdb_sql.parser import ast
from pandas import DataFrame

from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.libs.response import (
    HandlerResponse as Response,
)
from mindsdb.integrations.utilities.date_utils import parse_utc_date
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions


class AccountsTable(APITable):
    """
    Table class for the Google Content API for Shopping Accounts table.
    """

    def select(self, query: ast.Select) -> DataFrame:
        """
        Lists the sub-accounts in your Merchant Center account.

        Args:
            query (ast.Select): SQL query to parse.

        Returns:
            Response: Response object containing the results.
        """

        # Parse the query to get the conditions.
        conditions = extract_comparison_conditions(query.where)
        # Get the start and end times from the conditions.
        params = {}
        accepted_params = ['view', 'label', 'name']
        for op, arg1, arg2 in conditions:
            if arg1 == 'accountId':
                if op == '=':
                    params[arg1] = arg2
                elif op == '>':
                    params['startId'] = arg2
                elif op == '<':
                    params['endId'] = arg2
                else:
                    raise NotImplementedError
            if arg1 in accepted_params:
                if op != '=':
                    raise NotImplementedError
                params[arg1] = arg2
            else:
                raise NotImplementedError

        if query.limit is not None:
            params['maxResults'] = query.limit.value

        # Get the accounts from the API.
        accounts = self.handler. \
            call_application_api(method_name='get_accounts', params=params)

        selected_columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")

        if len(accounts) == 0:
            accounts = pd.DataFrame([], columns=selected_columns)
        else:
            accounts.columns = self.get_columns()
            for col in set(accounts.columns).difference(set(selected_columns)):
                accounts = accounts.drop(col, axis=1)
        return accounts

    def delete(self, query: ast.Delete):
        """
        Deletes accounts from your Merchant Center account.

        Args:
            query (ast.Delete): SQL query to parse.

        Returns:
            Response: Response object containing the results.
        """

        # Parse the query to get the conditions.
        conditions = extract_comparison_conditions(query.where)
        # Get the start and end times from the conditions.
        params = {}
        for op, arg1, arg2 in conditions:
            if arg1 == 'accountId':
                if op == '=':
                    params[arg1] = arg2
                elif op == '>':
                    params['startId'] = arg2
                elif op == '<':
                    params['endId'] = arg2
            elif arg1 == 'force':
                if op != '=':
                    raise NotImplementedError
                params[arg1] = arg2
            else:
                raise NotImplementedError

        # Delete the events in the Google Calendar API.
        self.handler.call_application_api(method_name='delete_accounts', params=params)

    def get_columns(self) -> list:
        """Gets all columns to be returned in pandas DataFrame responses"""
        return [
            'name',
            'kind',
            'websiteUrl',
            'adultContent',
            'sellerId',
            'users',
            'id',
            'youtubeChannelLinks',
            'googleMyBusinessLink',
            'businessInformation',
            'automaticImprovements',
            'adsLinks',
            'cssId',
            'labelIds',
            'accountManagement',
            'automaticLabelIds',
            'conversionSettings'
        ]


class OrdersTable(APITable):
    """
    Table class for the Google Content API for Shopping Orders table.
    """

    def select(self, query: ast.Select) -> DataFrame:
        """
        Lists the orders in your Merchant Center account.

        Args:
            query (ast.Select): SQL query to parse.

        Returns:
            Response: Response object containing the results.
        """

        # Parse the query to get the conditions.
        conditions = extract_comparison_conditions(query.where)
        # Get the start and end times from the conditions.
        params = {}
        accepted_params = ['statuses', 'acknowledged']
        for op, arg1, arg2 in conditions:
            if arg1 == 'orderId':
                if op == '=':
                    params[arg1] = arg2
                elif op == '>':
                    params['startId'] = arg2
                elif op == '<':
                    params['endId'] = arg2
                else:
                    raise NotImplementedError
            if arg1 == 'placedDateStart' or arg1 == 'placedDateEnd':
                if op != '=':
                    raise NotImplementedError
                params[arg1] = parse_utc_date(arg2)
            if arg1 in accepted_params:
                params[arg1] = parse_utc_date(arg2)
                if op != '=':
                    raise NotImplementedError
                params[arg1] = arg2
            else:
                raise NotImplementedError

        if query.order_by is not None:
            if query.order_by[0].value == 'placedDate':
                if query.order_by[1].value == 'ASC':
                    params['orderBy'] = 'placedDateAsc'
                else:
                    params['orderBy'] = 'placedDateDesc'
                raise NotImplementedError
            else:
                raise NotImplementedError

        if query.limit is not None:
            params['maxResults'] = query.limit.value

        # Get the orders from the API.
        orders = self.handler. \
            call_application_api(method_name='get_orders', params=params)

        selected_columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")

        if len(orders) == 0:
            orders = pd.DataFrame([], columns=selected_columns)
        else:
            orders.columns = self.get_columns()
            for col in set(orders.columns).difference(set(selected_columns)):
                orders = orders.drop(col, axis=1)
        return orders

    def delete(self, query: ast.Delete):
        """
        Deletes orders in your Merchant Center account.

        Args:
            query (ast.Delete): SQL query to parse.

        Returns:
            Response: Response object containing the results.
        """

        # Parse the query to get the conditions.
        conditions = extract_comparison_conditions(query.where)
        # Get the start and end times from the conditions.
        params = {}
        for op, arg1, arg2 in conditions:
            if arg1 == 'orderId':
                if op == '=':
                    params[arg1] = arg2
                elif op == '>':
                    params['startId'] = arg2
                elif op == '<':
                    params['endId'] = arg2
                else:
                    raise NotImplementedError
            else:
                raise NotImplementedError

        # Delete the events in the Google Calendar API.
        self.handler.call_application_api(method_name='delete_orders', params=params)

    def get_columns(self) -> list:
        """Gets all columns to be returned in pandas DataFrame responses"""
        return [
            'id',
            'merchantId',
            'merchantOrderId',
            'kind',
            'lineItems',
            'status',
            'paymentStatus',
            'acknowledged',
            'placedDate',
            'deliveryDetails',
            'customer',
            'shippingCost',
            'shippingCostTax',
            'refunds',
            'shipments',
            'billingAddress',
            'promotions',
            'taxCollector',
            'netPriceAmount',
            'netTaxAmount',
            'pickupDetails',
            'annotations'
        ]


class ProductsTable(APITable):
    """
    Table class for the Google Content API for Shopping Products table.
    """

    def select(self, query: ast.Select) -> DataFrame:
        """
        Lists the products in your Merchant Center account.

        Args:
            query (ast.Select): SQL query to parse.

        Returns:
            Response: Response object containing the results.
        """

        # Parse the query to get the conditions.
        conditions = extract_comparison_conditions(query.where)
        params = {}
        for op, arg1, arg2 in conditions:
            if arg1 == 'productId':
                if op == '=':
                    params[arg1] = arg2
                elif op == '>':
                    params['startId'] = arg2
                elif op == '<':
                    params['endId'] = arg2
                else:
                    raise NotImplementedError
            else:
                raise NotImplementedError

        if query.limit is not None:
            params['maxResults'] = query.limit.value

        # Get the products from the API.
        products = self.handler. \
            call_application_api(method_name='get_products', params=params)

        selected_columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")

        if len(products) == 0:
            products = pd.DataFrame([], columns=selected_columns)
        else:
            products.columns = self.get_columns()
            for col in set(products.columns).difference(set(selected_columns)):
                products = products.drop(col, axis=1)
        return products

    def update(self, query: ast.Update):
        """
        Updates products in your Merchant Center account.

        Args:
            query (ast.Update): SQL query to parse.

        Returns:
            Response: Response object containing the results.
        """

        params = {}
        values = query.values[0]
        # Get the event data from the values.
        accepted_params = self.get_columns()
        for col, val in zip(query.update_columns, values):
            if col in accepted_params:
                params[col] = val
            else:
                raise NotImplementedError

        params['updateMask'] = ','.join(params.keys())
        # Parse the query to get the conditions.
        conditions = extract_comparison_conditions(query.where)
        # Get the start and end times from the conditions.

        for op, arg1, arg2 in conditions:
            if arg1 == 'productId':
                if op == '=':
                    params[arg1] = arg2
                elif op == '>':
                    params['startId'] = arg2
                elif op == '<':
                    params['endId'] = arg2
                else:
                    raise NotImplementedError
            else:
                raise NotImplementedError

        # Update the products in the Google Merchant Center API.
        self.handler.call_application_api(method_name='update_products', params=params)

    def delete(self, query: ast.Delete):
        """
        Deletes products in your Merchant Center account.

        Args:
            query (ast.Delete): SQL query to parse.

        Returns:
            Response: Response object containing the results.
        """

        # Parse the query to get the conditions.
        conditions = extract_comparison_conditions(query.where)
        # Get the start and end times from the conditions.
        params = {}
        for op, arg1, arg2 in conditions:
            if arg1 == 'productId':
                if op == '=':
                    params[arg1] = arg2
                elif op == '>':
                    params['startId'] = arg2
                elif op == '<':
                    params['endId'] = arg2
                else:
                    raise NotImplementedError
            elif arg1 == 'feedId':
                if op != '=':
                    raise NotImplementedError
                params[arg1] = arg2
            else:
                raise NotImplementedError

        # Delete the products in the Google Merchant Center API.
        self.handler.call_application_api(method_name='delete_products', params=params)

    def get_columns(self) -> list:
        """Gets all columns to be returned in pandas DataFrame responses"""
        return [
            'id',
            'offerId',
            'title',
            'description',
            'link',
            'imageLink',
            'contentLanguage',
            'targetCountry',
            'channel',
            'channelExclusivity',
            'price',
            'salePrice',
            'salePriceEffectiveDate',
            'gtin',
            'mpn',
            'brand',
            'condition',
            'adult',
            'multipack',
            'isBundle',
            'energyEfficiencyClass',
            'minEnergyEfficiencyClass',
            'maxEnergyEfficiencyClass',
            'ageGroup',
            'color',
            'expirationDate',
            'disclosureDate',
            'availability',
            'source'
        ]



import argparse
import datetime
from functools import wraps
import hashlib
import base64
from cryptography.fernet import Fernet
from collections.abc import Callable

import requests
from mindsdb_sql import get_lexer_parser
from mindsdb_sql.parser.ast import Identifier

from mindsdb.utilities.fs import create_process_mark, delete_process_mark, set_process_mark


def args_parse():
    parser = argparse.ArgumentParser(description='CL argument for mindsdb server')
    parser.add_argument('--api', type=str, default=None)
    parser.add_argument('--config', type=str, default=None)
    parser.add_argument('--install-handlers', type=str, default=None)
    parser.add_argument('--verbose', action='store_true')
    parser.add_argument('--no_studio', action='store_true')
    parser.add_argument('-v', '--version', action='store_true')
    parser.add_argument('--ml_task_queue_consumer', action='store_true', default=None)
    return parser.parse_args()


def cast_row_types(row, field_types):
    '''
    '''
    keys = [x for x in row.keys() if x in field_types]
    for key in keys:
        t = field_types[key]
        if t == 'Timestamp' and isinstance(row[key], (int, float)):
            timestamp = datetime.datetime.utcfromtimestamp(row[key])
            row[key] = timestamp.strftime('%Y-%m-%d %H:%M:%S')
        elif t == 'Date' and isinstance(row[key], (int, float)):
            timestamp = datetime.datetime.utcfromtimestamp(row[key])
            row[key] = timestamp.strftime('%Y-%m-%d')
        elif t == 'Int' and isinstance(row[key], (int, float, str)):
            try:
                print(f'cast {row[key]} to {int(row[key])}')
                row[key] = int(row[key])
            except Exception:
                pass


def is_notebook():
    try:
        if 'IPKernelApp' in get_ipython().config:
            return True
        else:
            return False
    except NameError:
        return False      # Probably standard Python interpreter


def mark_process(name: str, custom_mark: str = None) -> Callable:
    def mark_process_wrapper(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args, **kwargs):
            if custom_mark is None:
                mark = create_process_mark(name)
            else:
                mark = set_process_mark(name, custom_mark)

            try:
                return func(*args, **kwargs)
            finally:
                delete_process_mark(name, mark)
        return wrapper
    return mark_process_wrapper


def get_versions_where_predictors_become_obsolete():
    """ Get list of MindsDB versions in which predictors should be retrained
        Returns:
            list of str or False
    """
    versions_for_updating_predictors = []
    try:
        try:
            res = requests.get(
                'https://mindsdb-cloud-public-service-files.s3.us-east-2.amazonaws.com/version_for_updating_predictors.txt',
                timeout=0.5
            )
        except (ConnectionError, requests.exceptions.ConnectionError) as e:
            print(f'Is no connection. {e}')
            raise
        except Exception as e:
            print(f'Is something wrong with getting version_for_updating_predictors.txt: {e}')
            raise

        if res.status_code != 200:
            print(f'Cant get version_for_updating_predictors.txt: returned status code = {res.status_code}')
            raise

        try:
            versions_for_updating_predictors = res.text.replace(' \t\r', '').split('\n')
        except Exception as e:
            print(f'Cant decode version_for_updating_predictors.txt: {e}')
            raise
    except Exception:
        return False, versions_for_updating_predictors

    versions_for_updating_predictors = [x for x in versions_for_updating_predictors if len(x) > 0]
    return True, versions_for_updating_predictors


def init_lexer_parsers():
    get_lexer_parser('mindsdb')
    get_lexer_parser('mysql')


def resolve_model_identifier(name: Identifier) -> tuple:
    """ split model name to parts

        Examples:
            >>> resolve_model_identifier(['a', 'b'])
            ('a', 'b', None)

            >>> resolve_model_identifier(['a', '1'])
            (None, 'a', 1)

            >>> resolve_model_identifier(['a'])
            (None, 'a', None)

            >>> resolve_model_identifier(['a', 'b', 'c'])
            (None, None, None)  # not found

        Args:
            name (list): Identifier parts

        Returns:
            tuple: (database_name, model_name, model_version, describe)
    """
    name = name.parts
    database_name = None
    model_name = None
    model_version = None
    parts_count = len(name)
    if parts_count == 1:
        database_name = None
        model_name = name[0]
        model_version = None
    elif parts_count == 2:
        if name[-1].isdigit():
            database_name = None
            model_name = name[0]
            model_version = int(name[-1])
        else:
            database_name = name[0]
            model_name = name[1]
            model_version = None
    elif parts_count == 3:
        database_name = name[0]
        model_name = name[1]
        if name[2].isdigit():
            model_version = int(name[2])
        else:
            # not found
            return None, None, None

    return database_name, model_name, model_version


def encrypt(string: bytes, key: str) -> bytes:
    hashed_string = hashlib.sha256(key.encode()).digest()

    fernet_key = base64.urlsafe_b64encode(hashed_string)

    cipher = Fernet(fernet_key)
    return cipher.encrypt(string)


def decrypt(encripted: bytes, key: str) -> bytes:
    hashed_string = hashlib.sha256(key.encode()).digest()

    fernet_key = base64.urlsafe_b64encode(hashed_string)

    cipher = Fernet(fernet_key)
    return cipher.decrypt(encripted)



"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""

from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packet import Packet
from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import CHARSET_NUMBERS, TYPES
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_datum import Datum


class ColumnDefenitionPacket(Packet):
    '''
    Implementation based on:
    https://mariadb.com/kb/en/library/resultset/
    '''

    # https://dev.mysql.com/doc/internals/en/com-query-response.html
    def setup(self):
        self.catalog = Datum('string<lenenc>', 'def')
        self.schema = Datum(
            'string<lenenc>',
            self._kwargs.get('schema', ''))
        self.table_alias = Datum(
            'string<lenenc>',
            self._kwargs.get('table_alias', '')
        )
        self.table_name = Datum(
            'string<lenenc>',
            self._kwargs.get('table_name', '')
        )
        self.column_alias = Datum(
            'string<lenenc>',
            self._kwargs.get('column_alias', '')
        )
        self.column_name = Datum(
            'string<lenenc>',
            self._kwargs.get('column_name', '')
        )
        self.fixed_length = Datum('int<lenenc>', 0xC)
        charset = self._kwargs.get('charset', CHARSET_NUMBERS["utf8_unicode_ci"])
        self.character_set = Datum('int<2>', charset)
        self.column_length = Datum('int<4>', self._kwargs.get('max_length', 0xf))  # may be this? https://books.google.ru/books?id=G2YqBS9CQ0AC&lpg=PP1&hl=ru&pg=PA428#v=onepage&q&f=false
        self.column_type = Datum(
            'int<1>',
            self._kwargs.get(
                'column_type',
                self._kwargs.get('column_type', TYPES.MYSQL_TYPE_VARCHAR)
            )
        )

        self.flags = Datum('int<2>', self._kwargs.get('flags', 0))
        self.decimals = Datum('int<1>', 0)

        self.unused = Datum('int<2>', 0)

    @property
    def body(self):
        order = [
            'catalog',
            'schema',
            'table_alias',
            'table_name',
            'column_alias',
            'column_name',
            'fixed_length',
            'character_set',
            'column_length',
            'column_type',
            'flags',
            'decimals',
            'unused'
        ]
        string = b''
        for key in order:
            string += getattr(self, key).toStringPacket()

        self.setBody(string)
        return self._body

    @staticmethod
    def test():
        import pprint
        pprint.pprint(str(ColumnDefenitionPacket().get_packet_string()))


# only run the test if this file is called from debugger
if __name__ == "__main__":
    ColumnDefenitionPacket.test()



import pickle

from walrus import Database
from pandas import DataFrame

from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.config import Config
from mindsdb.utilities.ml_task_queue.utils import RedisKey, to_bytes
from mindsdb.utilities.ml_task_queue.task import Task
from mindsdb.utilities.ml_task_queue.base import BaseRedisQueue
from mindsdb.utilities.ml_task_queue.const import (
    TASKS_STREAM_NAME,
    ML_TASK_TYPE,
    ML_TASK_STATUS
)


class MLTaskProducer(BaseRedisQueue):
    """ Interface around the redis for putting tasks to the queue

        Attributes:
            db (Redis): database object
            stream
            cache
            pubsub
    """

    def __init__(self) -> None:
        config = Config().get('ml_task_queue', {})

        self.db = Database(
            host=config.get('host', 'localhost'),
            port=config.get('port', 6379),
            db=config.get('db', 0),
            username=config.get('username'),
            password=config.get('password'),
            protocol=3
        )
        self.wait_redis_ping(60)

        self.stream = self.db.Stream(TASKS_STREAM_NAME)
        self.cache = self.db.cache()
        self.pubsub = self.db.pubsub()

    def apply_async(self, task_type: ML_TASK_TYPE, model_id: int, payload: dict, dataframe: DataFrame = None) -> Task:
        ''' Add tasks to the queue

            Args:
                task_type (ML_TASK_TYPE): type of the task
                model_id (int): model identifier
                payload (dict): lightweight model data that will be added to stream message
                dataframe (DataFrame): dataframe will be transfered via regular redis storage

            Returns:
                Task: object representing the task
        '''
        try:
            payload = pickle.dumps(payload, protocol=5)
            redis_key = RedisKey.new()
            message = {
                "task_type": task_type.value,
                "company_id": '' if ctx.company_id is None else ctx.company_id,     # None can not be dumped
                "model_id": model_id,
                "payload": payload,
                "redis_key": redis_key.base
            }

            self.wait_redis_ping()
            if dataframe is not None:
                self.cache.set(redis_key.dataframe, to_bytes(dataframe), 180)
            self.cache.set(redis_key.status, ML_TASK_STATUS.WAITING, 180)

            self.stream.add(message)
            return Task(self.db, redis_key)
        except ConnectionError:
            print('Cant send message to redis: connect failed')
            raise



import numpy as np
import pandas as pd
from pandas.tseries.frequencies import to_offset
from sklearn.metrics import r2_score
from hierarchicalforecast.core import HierarchicalReconciliation
from hierarchicalforecast.methods import BottomUp
from hierarchicalforecast.utils import aggregate

DEFAULT_FREQUENCY = "D"
DEFAULT_RECONCILER = BottomUp


def transform_to_nixtla_df(df, settings_dict, exog_vars=[]):
    """Transform dataframes into the specific format required by StatsForecast.

    Nixtla packages require dataframes to have the following columns:
        unique_id -> the grouping column. If multiple groups are specified then
        we join them into one name using a / char.
        ds -> the date series
        y -> the target variable for prediction

    You can optionally include exogenous regressors after these three columns, but
    they must be numeric.
    """
    nixtla_df = df.copy()

    # Resample every group
    freq = settings_dict['frequency']
    resampled_df = pd.DataFrame(columns=nixtla_df.columns)
    if settings_dict["group_by"] and settings_dict["group_by"] != ['__group_by']:
        for group, groupdf in nixtla_df.groupby(by=settings_dict["group_by"]):
            groupdf.index = pd.to_datetime(groupdf.pop(settings_dict["order_by"]))
            resampled_groupdf = pd.DataFrame(groupdf[settings_dict['target']].resample(freq).mean())
            for k, v in zip(settings_dict["group_by"], group):
                resampled_groupdf[k] = v
            resampled_groupdf = resampled_groupdf.reset_index()
            resampled_df = pd.concat([resampled_df, resampled_groupdf])
        nixtla_df = resampled_df

    # Transform group columns into single unique_id column
    if len(settings_dict["group_by"]) > 1:
        for col in settings_dict["group_by"]:
            nixtla_df[col] = nixtla_df[col].astype(str)
        nixtla_df["unique_id"] = nixtla_df[settings_dict["group_by"]].agg("/".join, axis=1)
        group_col = "ignore this"
    else:
        group_col = settings_dict["group_by"][0]

    # Rename columns to statsforecast names
    nixtla_df = nixtla_df.rename(
        {settings_dict["target"]: "y", settings_dict["order_by"]: "ds", group_col: "unique_id"}, axis=1
    )

    if "unique_id" not in nixtla_df.columns:
        # add to dataframe as it is expected by statsforecast
        nixtla_df["unique_id"] = '1'

    columns_to_keep = ["unique_id", "ds", "y"] + exog_vars
    nixtla_df["ds"] = pd.to_datetime(nixtla_df["ds"])
    return nixtla_df[columns_to_keep]


def get_results_from_nixtla_df(nixtla_df, model_args):
    """Transform dataframes generated by StatsForecast back to their original format.

    This will return the dataframe to the original format supplied by the MindsDB query.
    """
    return_df = nixtla_df.reset_index(drop=True if 'unique_id' in nixtla_df.columns else False)
    if len(model_args["group_by"]) > 0:
        if len(model_args["group_by"]) > 1:
            for i, group in enumerate(model_args["group_by"]):
                return_df[group] = return_df["unique_id"].apply(lambda x: x.split("/")[i])
        else:
            group_by_col = model_args["group_by"][0]
            return_df[group_by_col] = return_df["unique_id"]

    return return_df.drop(["unique_id"], axis=1).rename({"ds": model_args["order_by"]}, axis=1)


def infer_frequency(df, time_column, default=DEFAULT_FREQUENCY):
    try:  # infer frequency from time column
        date_series = pd.to_datetime(df.sort_values(by=time_column)[time_column]).unique()
        inferred_freq = pd.infer_freq(date_series)  # call this first to get e.g. months & other irregular periods right
        if inferred_freq is None:
            values, counts = np.unique(np.diff(date_series), return_counts=True)
            delta = values[np.argmax(counts)]
            inferred_freq = to_offset(pd.to_timedelta(delta)).freqstr
    except TypeError:
        inferred_freq = default
    return inferred_freq if inferred_freq is not None else default


def get_model_accuracy_dict(nixtla_results_df, metric=r2_score):
    """Calculates accuracy for each model in the nixtla results df."""
    accuracy_dict = {}
    for column in nixtla_results_df.columns:
        if column in ["unique_id", "ds", "y", "cutoff"]:
            continue
        model_error = metric(nixtla_results_df["y"], nixtla_results_df[column])
        accuracy_dict[column] = model_error
    return accuracy_dict


def get_best_model_from_results_df(nixtla_results_df, metric=r2_score):
    """Gets the best model based, on lowest error, from a results df
    with a column for each nixtla model.
    """
    best_model, current_accuracy = None, 0
    accuracy_dict = get_model_accuracy_dict(nixtla_results_df, metric)
    for model, accuracy in accuracy_dict.items():
        if accuracy > current_accuracy:
            best_model, current_accuracy = model, accuracy
    return best_model


def spec_hierarchy_from_list(col_list):
    """Gets the hierarchy spec from the list of hierarchy cols"""
    spec = [["Total"]]
    for i in range(len(col_list)):
        spec.append(["Total"] + col_list[: i + 1])
    return spec


def get_hierarchy_from_df(df, model_args):
    """Extracts hierarchy from the raw df, using the provided spec and args.

    The "hierarchy" model arg is a list of format
    [<level 1>, <level 2>, ..., <level n>]
    where each element is a level in the hierarchy.

    We return a tuple (nixtla_df, hier_df, hier_dict) where:
    nixtla_df is a dataframe in the format nixtla packages uses for training
    hier_df is a matrix of 0s and 1s showing the hierarchical structure
    hier_dict is a dictionary with the hierarchical structure. See the unit test
    in tests/unit/ml_handlers/test_time_series_utils.py for an example.
    """
    spec = spec_hierarchy_from_list(model_args["hierarchy"])

    nixtla_df = df.rename({model_args["order_by"]: "ds", model_args["target"]: "y"}, axis=1)
    nixtla_df["ds"] = pd.to_datetime(nixtla_df["ds"])
    for col in model_args["group_by"]:
        nixtla_df[col] = nixtla_df[col].astype(str)  # grouping columns need to be string format
    nixtla_df.insert(0, "Total", "total")

    nixtla_df, hier_df, hier_dict = aggregate(nixtla_df, spec)  # returns (nixtla_df, hierarchy_df, hierarchy_dict)
    return nixtla_df, hier_df, hier_dict


def reconcile_forecasts(nixtla_df, forecast_df, hierarchy_df, hierarchy_dict):
    """Reconciles forecast results according to the hierarchy."""
    reconcilers = [DEFAULT_RECONCILER()]
    hrec = HierarchicalReconciliation(reconcilers=reconcilers)
    reconciled_df = hrec.reconcile(Y_hat_df=forecast_df, Y_df=nixtla_df, S=hierarchy_df, tags=hierarchy_dict)
    return get_results_from_reconciled_df(reconciled_df, hierarchy_df)


def get_results_from_reconciled_df(reconciled_df, hierarchy_df):
    """Formats the reconciled df into a normal Nixtla results df.

    First drops the model output columns that haven't been reconciled.
    Then drops rows corresponding to higher level predictions that were not
    in the original dataframe, e.g. the total for each grouping.
    """
    #  Drop unnecessary columns
    for col in reconciled_df.columns:
        if col not in ["ds", "y"]:
            if "BottomUp" not in col:
                results_df = reconciled_df.drop(col, axis=1)  # removes original forecast column
                break

    #  Drop higher-level rows
    lowest_level_ids = hierarchy_df.columns
    results_df = results_df[results_df.index.isin(lowest_level_ids)]
    results_df.index = results_df.index.str.replace("total/", "")
    return results_df



from typing import List
import pandas as pd
from mindsdb.integrations.libs.api_handler import APIHandler, APITable
from mindsdb_sql.parser import ast
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb_sql.parser.ast.select.constant import Constant
import json


class StrapiTable(APITable):

    def __init__(self, handler: APIHandler, name: str):
        super().__init__(handler)
        self.name = name
        # get all the fields of a collection as columns
        self.columns = self.handler.call_strapi_api(method='GET', endpoint=f'/api/{name}').columns

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Triggered at the SELECT query

        Args:
            query (ast.Select): User's entered query

        Returns:
            pd.DataFrame: The queried information
        """
        # Initialize _id and selected_columns
        _id = None
        selected_columns = []

        # Get id from where clause, if available
        conditions = extract_comparison_conditions(query.where)
        for op, arg1, arg2 in conditions:
            if arg1 == 'id' and op == '=':
                _id = arg2
            else:
                raise ValueError("Unsupported condition in WHERE clause")

        # Get selected columns from query
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")

        # Initialize the result DataFrame
        result_df = None

        if _id is not None:
            # Fetch data using the provided endpoint for the specific id
            df = self.handler.call_strapi_api(method='GET', endpoint=f'/api/{self.name}/{_id}')

            if len(df) > 0:
                result_df = df[selected_columns]
        else:
            # Fetch data without specifying an id
            page_size = 100  # The page size you want to use for API requests
            limit = query.limit.value if query.limit else None
            result_df = pd.DataFrame(columns=selected_columns)

            if limit:
                # Calculate the number of pages required
                page_count = (limit + page_size - 1) // page_size
            else:
                page_count = 1

            for page in range(1, page_count + 1):
                if limit:
                    # Calculate the page size for this request
                    current_page_size = min(page_size, limit)
                else:
                    current_page_size = page_size

                df = self.handler.call_strapi_api(method='GET', endpoint=f'/api/{self.name}', params={'pagination[page]': page, 'pagination[pageSize]': current_page_size})

                if len(df) == 0:
                    break

                result_df = pd.concat([result_df, df[selected_columns]], ignore_index=True)

                if limit:
                    limit -= current_page_size

        return result_df

    def insert(self, query: ast.Insert) -> None:
        """triggered at the INSERT query
        Args:
            query (ast.Insert): user's entered query
        """
        data = {'data': {}}
        for column, value in zip(query.columns, query.values[0]):
            if isinstance(value, Constant):
                data['data'][column.name] = value.value
            else:
                data['data'][column.name] = value
        self.handler.call_strapi_api(method='POST', endpoint=f'/api/{self.name}', json_data=json.dumps(data))

    def update(self, query: ast.Update) -> None:
        """triggered at the UPDATE query

        Args:
            query (ast.Update): user's entered query
        """
        conditions = extract_comparison_conditions(query.where)
        # Get id from query
        for op, arg1, arg2 in conditions:
            if arg1 == 'id' and op == '=':
                _id = arg2
            else:
                raise NotImplementedError
        data = {'data': {}}
        for key, value in query.update_columns.items():
            if isinstance(value, Constant):
                data['data'][key] = value.value
        self.handler.call_strapi_api(method='PUT', endpoint=f'/api/{self.name}/{_id}', json_data=json.dumps(data))

    def get_columns(self, ignore: List[str] = []) -> List[str]:
        """columns

        Args:
            ignore (List[str], optional): exclusion items. Defaults to [].

        Returns:
            List[str]: available columns with `ignore` items removed from the list.
        """

        return [item for item in self.columns if item not in ignore]
